part of arasu.resource;

class Store {
  String name,origin,path;
  Db db;
  Type type;
  Http http;
  bool _once=true;
  Future _opened;
  List _caches = new List();
  ClassMirror cm;

  Store(this.type,this.db,this.http,this.name,this.origin,this.path){ 
    cm = reflectClass(type);
  }
  
  Future get opened {
    if (!db.dbase.objectStoreNames.contains(name)) {
      //Logger.root.info('$name store does not exists, so creating it');
      //Logger.root.info('$name store opened sucessfully ');
      db.dbase.close();
      _opened = window.indexedDB.open(db.dbase.name, version: db.dbase.version.hashCode+1, 
          onUpgradeNeeded: (VersionChangeEvent e) =>
              ((e.target as Request).result as Database).createObjectStore(name, autoIncrement: true)
        ).then((e) => db.dbase = e);
    }else {
      _opened = new Future.value();
    }
    return _opened;
  }
  Future load() {
    var transaction = db.dbase.transaction(name, 'readwrite');
    var objectStore = transaction.objectStore(name);
    var cursors = objectStore.openCursor(autoAdvance: true).asBroadcastStream();
    cursors.listen((cursor) {
      var obj = cm.newInstance(new Symbol(''),[],{}).reflectee;
      Serializer.decode(cursor.value , obj);
      syncCache(obj);
    });
    return new Future.value();
  }
  
  List get caches => _caches;
 
//  List get caches { 
//    if(_caches==null) {
//      _caches = new List();
//      load();
//      fetch();
//    }
//    return _caches;
//  }
  //InstanceMirror im = cm.newInstance(new Symbol(''),[],{}).reflectee;
Future insert(List<Map<String,dynamic>> records){
   var transaction = db.dbase.transaction(name, 'readwrite');
  var objectStore = transaction.objectStore(name);
  records.forEach((e){
    var key = e['Id'];
    objectStore.getObject(key).then((map){
      if(map==null){
        objectStore.add(e,key).then((addedKey) {
          var obj = cm.newInstance(new Symbol(''),[],{}).reflectee;
          Serializer.decode(e, obj);
          _caches.add(obj);
          //obj.Id = addedKey;
        });
      }else if(map.toString()!=e.toString()) {
        var obj = _caches.firstWhere((e)=>e.Id==key,orElse:()=> null);
        Serializer.decode(e, obj);
        objectStore.put(e,obj.Id).then((addedKey) {
          //obj.Id = addedKey;
        });
      }
    }).catchError(Logger.root.shout);
  });
  return transaction.completed;

}
void once() {
  if(_once){
    load();
    fetch();
    _once=false;
    
  }
}
  Future fetch(){
    return http.get(path).then((HttpResponse response) {
      List data = response.data;
      if (data[1]==null && data[0]!=null && (data[0] as List).length>0) {
        return insert(data[0]);
      }else {
        Logger.root.shout("Server Response Error :",JSON.decode(data[1]));
      }
    });
  }

  Future remove(dynamic obj) {
    return http.delete('$path/${obj.Id}').then((HttpResponse response) {
      List data = response.data;
      if (data[0]==null) {
        var transaction = db.dbase.transaction(name, 'readwrite');
        transaction.objectStore(name).delete(obj.Id)
        ..then((_)=>_caches.remove(obj))
        ..catchError(Logger.root.shout);
        return transaction.completed;
      }else {
        Logger.root.shout(JSON.decode(data[0]));
      }
    }).catchError(Logger.root.shout);    
  }
  dynamic find(dynamic key,{bool load:false,fetch:false}) {
    var obj;
    if (fetch || load) {
      obj = cm.newInstance(new Symbol(''),[],{}).reflectee;
      if(fetch) {
        findThroughFetch(key,obj);
        return obj;
      }else if(load){
        findThroughLoad(key,obj).catchError(Logger.root.shout);
        return obj;
      }
    }
    obj = _caches.firstWhere((e)=>e.Id==key,orElse:()=> null);
    if(obj!=null) return obj;
    else obj = cm.newInstance(new Symbol(''),[],{}).reflectee;
    findThroughLoad(key,obj).catchError(Logger.root.shout);
    return obj;
  }
  
  Future findThroughLoad(dynamic key,obj) {
    var transaction = db.dbase.transaction(name, 'readwrite');
    return transaction.objectStore(name).getObject(key).then((map){
      if(map!=null){
        Serializer.decode(map,obj);
        syncCache(obj);
      }else{
        return findThroughFetch(key,obj);
      }
    }).catchError((_){
      return findThroughFetch(key,obj);
    });
  }

  Future findThroughFetch(dynamic key, obj) {
    return http.get('$path/${key}').then((HttpResponse response) {
      List data = response.data;
      if (data[1]==null) {
        Serializer.decode(data[0],obj);
        var transaction = db.dbase.transaction(name, 'readwrite');
        transaction.objectStore(name).put(data[0],key)
        ..then((_)=> syncCache(obj) )
        ..catchError(Logger.root.shout);
        return transaction.completed;
      } else {
        Logger.root.shout("Server Response Error :",JSON.decode(data[1]));
      }
    }).catchError(Logger.root.shout);
  }
  void syncCache(var obj){
    var cacheObj = _caches.firstWhere((e)=>e.Id==obj.Id , orElse:()=> null);
    if(cacheObj!=null) {
      if(cacheObj== obj) return;
      Map a=cacheObj.encode(),b=obj.encode();
      Map changes = Serializer.subtractMapChanges(a,b);
      if (changes.isNotEmpty) Serializer.decode(changes, cacheObj);
    }else{
      _caches.add(obj);
    }
  }
 
  Future clear() {
    var transaction = db.dbase.transaction(name, 'readwrite');
    var objectStore = transaction.objectStore(name);
    objectStore.clear()
    ..then((_)=>_caches.clear())
    ..catchError(Logger.root.shout);
    return transaction.completed;
  }

  Future add(Map attrs,{var obj}) {
    if (obj==null) obj = cm.newInstance(new Symbol(''),[],{}).reflectee;
    return http.post(path,JSON.encode(attrs)).then((HttpResponse response) {
      List data = response.data;
      if (data[1]==null) {
        Serializer.decode(data[0], obj);
        var transaction = db.dbase.transaction(name, 'readwrite');
        transaction.objectStore(name).add(data[0],obj.Id)
        ..then((_)=>syncCache(obj))
        ..catchError(Logger.root.shout);
        return transaction.completed;
      }else {
        Logger.root.shout("Server Response Error :",JSON.decode(data[1]));
        return obj; //TODO : add error into obj 
      }
    }).catchError(Logger.root.shout);
  }  
  
  Future put(var obj,var attrs) {
    return http.put('$path/${obj.Id}', attrs).then((HttpResponse response) {
      List data = response.data;
      if (data[1]==null) {
        var transaction = db.dbase.transaction(name, 'readwrite');
        transaction.objectStore(name).put(data[0],obj.Id)
        ..then((addedKey) {
          //obj.Id = addedKey;
          syncCache(obj);
        })
        ..catchError(Logger.root.shout);
        return transaction.completed;
      } else {
        Logger.root.shout("Server Response Error :",JSON.decode(data[1]));
        return obj; //TODO : add error into obj 
      }
    }).catchError(Logger.root.shout);
  }  
}
